@@ -4,7 +4,7 @@ This module takes care of steps 3 and 4 of “CSS 2.1 processing model”: Retri
 stylesheets associated with a document and annotate every element with a value
 for every CSS property.
 
-http://www.w3.org/TR/CSS21/intro.html#processing-model
+https://www.w3.org/TR/CSS21/intro.html#processing-model
 
 This module does this in more than two steps. The
 :func:`get_all_computed_styles` function does everything, but it is itsef based
@@ -48,7 +48,7 @@ class StyleFor:
         #     values: (values, weight)
         #         values: a PropertyValue-like object
         #         weight: values with a greater weight take precedence, see
-        #             http://www.w3.org/TR/CSS21/cascade.html#cascading-order
+        #             https://www.w3.org/TR/CSS21/cascade.html#cascading-order
         self._cascaded_styles = cascaded_styles = {}
 
         # keys: (element, pseudo_element_type), like cascaded_styles
@@ -123,13 +123,13 @@ class StyleFor:
             if ('table' in style['display'] and
                     style['border_collapse'] == 'collapse'):
                 # Padding do not apply
-                for side in ['top', 'bottom', 'left', 'right']:
+                for side in ('top', 'bottom', 'left', 'right'):
                     style[f'padding_{side}'] = computed_values.ZERO_PIXELS
             if (len(style['display']) == 1 and
                     style['display'][0].startswith('table-') and
                     style['display'][0] != 'table-caption'):
                 # Margins do not apply
-                for side in ['top', 'bottom', 'left', 'right']:
+                for side in ('top', 'bottom', 'left', 'right'):
                     style[f'margin_{side}'] = computed_values.ZERO_PIXELS
 
         return style
@@ -211,12 +211,11 @@ class StyleFor:
         if selector_page_type.index is not None:
             a, b, group = selector_page_type.index
             # TODO: handle group
-            if a:
-                if (page_type.index + 1 - b) % a:
-                    return False
+            offset = page_type.index + 1 - b
+            if a == 0:
+                return offset == 0
             else:
-                if page_type.index + 1 != b:
-                    return False
+                return offset / a >= 0 and not offset % a
         return True
 
 
@@ -587,7 +586,7 @@ def declaration_precedence(origin, importance):
     and ``'user agent'``.
 
     """
-    # See http://www.w3.org/TR/CSS21/cascade.html#cascading-order
+    # See https://www.w3.org/TR/CSS21/cascade.html#cascading-order
     if origin == 'user agent':
         return 1
     elif origin == 'user' and not importance:
@@ -627,18 +626,18 @@ class AnonymousStyle(dict):
         return copy
 
     def __missing__(self, key):
-        if key in INHERITED or key.startswith('__'):
-            self[key] = self.parent_style[key]
+        if key in INHERITED or key[:2] == '__':
+            value = self[key] = self.parent_style[key]
         elif key == 'page':
             # page is not inherited but taken from the ancestor if 'auto'
-            self[key] = self.parent_style[key]
-        elif key.startswith('text_decoration_'):
-            self[key] = text_decoration(
+            value = self[key] = self.parent_style[key]
+        elif key[:16] == 'text_decoration_':
+            value = self[key] = text_decoration(
                 key, INITIAL_VALUES[key], self.parent_style[key],
                 cascaded=False)
         else:
-            self[key] = INITIAL_VALUES[key]
-        return self[key]
+            value = self[key] = INITIAL_VALUES[key]
+        return value
 
 
 class ComputedStyle(dict):
@@ -677,7 +676,7 @@ class ComputedStyle(dict):
         if key in self.cascaded:
             value = keyword = self.cascaded[key][0]
         else:
-            if key in INHERITED or key.startswith('__'):
+            if key in INHERITED or key[:2] == '__':
                 keyword = 'inherit'
             else:
                 keyword = 'initial'
@@ -687,7 +686,7 @@ class ComputedStyle(dict):
             keyword = 'initial'
 
         if keyword == 'initial':
-            value = None if key.startswith('__') else INITIAL_VALUES[key]
+            value = None if key[:2] == '__' else INITIAL_VALUES[key]
             if key not in INITIAL_NOT_COMPUTED:
                 # The value is the same as when computed
                 self[key] = value
@@ -695,20 +694,22 @@ class ComputedStyle(dict):
             # Values in parent_style are already computed.
             self[key] = value = self.parent_style[key]
 
-        if key.startswith('text_decoration_') and self.parent_style:
-            self[key] = text_decoration(
+        if key[:16] == 'text_decoration_' and self.parent_style:
+            value = text_decoration(
                 key, value, self.parent_style[key], key in self.cascaded)
-
-        if key == 'page' and value == 'auto':
+            if key in self:
+                del self[key]
+        elif key == 'page' and value == 'auto':
             # The page property does not inherit. However, if the page
             # value on an element is auto, then its used value is the value
             # specified on its nearest ancestor with a non-auto value. When
             # specified on the root element, the used value for auto is the
             # empty string.
-            self['page'] = value = (
+            value = (
                 '' if self.parent_style is None else self.parent_style['page'])
-
-        if key in ('position', 'float', 'display'):
+            if key in self:
+                del self[key]
+        elif key in ('position', 'float', 'display'):
             self.specified[key] = value
 
         if key in self:
@@ -749,7 +750,7 @@ def computed_from_cascaded(element, cascaded, parent_style, pseudo_type=None,
 
     style = ComputedStyle(
         parent_style, cascaded, element, pseudo_type, root_style, base_url)
-    if target_collector:
+    if target_collector and style['anchor']:
         target_collector.collect_anchor(style['anchor'])
     return style
 
@@ -874,12 +875,9 @@ def parse_page_selectors(rule):
 
 
 def preprocess_stylesheet(device_media_type, base_url, stylesheet_rules,
-                          url_fetcher, matcher, page_rules, fonts,
-                          font_config, counter_style, ignore_imports=False):
-    """Do the work that can be done early on stylesheet, before they are
-    in a document.
-
-    """
+                          url_fetcher, matcher, page_rules, font_config,
+                          counter_style, ignore_imports=False):
+    """Do what can be done early on stylesheet, before being in a document."""
     for rule in stylesheet_rules:
         if getattr(rule, 'content', None) is None and (
                 rule.type != 'at-rule' or rule.lower_at_keyword != 'import'):
@@ -975,7 +973,7 @@ def preprocess_stylesheet(device_media_type, base_url, stylesheet_rules,
             content_rules = tinycss2.parse_rule_list(rule.content)
             preprocess_stylesheet(
                 device_media_type, base_url, content_rules, url_fetcher,
-                matcher, page_rules, fonts, font_config, counter_style,
+                matcher, page_rules, font_config, counter_style,
                 ignore_imports=True)
 
         elif rule.type == 'at-rule' and rule.lower_at_keyword == 'page':
@@ -1028,10 +1026,7 @@ def preprocess_stylesheet(device_media_type, base_url, stylesheet_rules,
                     break
             else:
                 if font_config is not None:
-                    font_filename = font_config.add_font_face(
-                        rule_descriptors, url_fetcher)
-                    if font_filename:
-                        fonts.append(font_filename)
+                    font_config.add_font_face(rule_descriptors, url_fetcher)
 
         elif (rule.type == 'at-rule' and
                 rule.lower_at_keyword == 'counter-style'):
@@ -1101,7 +1096,7 @@ def preprocess_stylesheet(device_media_type, base_url, stylesheet_rules,
 def get_all_computed_styles(html, user_stylesheets=None,
                             presentational_hints=False, font_config=None,
                             counter_style=None, page_rules=None,
-                            target_collector=None):
+                            target_collector=None, forms=False):
     """Compute all the computed styles of all elements in ``html`` document.
 
     Do everything from finding author stylesheets to parsing and applying them.
@@ -1117,7 +1112,7 @@ def get_all_computed_styles(html, user_stylesheets=None,
     for style in html._ua_counter_style():
         for key, value in style.items():
             counter_style[key] = value
-    for sheet in (html._ua_stylesheets() or []):
+    for sheet in (html._ua_stylesheets(forms) or []):
         sheets.append((sheet, 'user agent', None))
     if presentational_hints:
         for sheet in (html._ph_stylesheets() or []):
