@@ -78,10 +78,12 @@ def draw_box_background_and_border(stream, page, box):
 
 def draw_stacking_context(stream, stacking_context):
     """Draw a ``stacking_context`` on ``stream``."""
-    # See http://www.w3.org/TR/CSS2/zindex.html
+    # See https://www.w3.org/TR/CSS2/zindex.html
     with stacked(stream):
         box = stacking_context.box
 
+        stream.begin_marked_content(box, mcid=True)
+
         # apply the viewport_overflow to the html box, see #35
         if box.is_for_root_element and (
                 stacking_context.page.style['overflow'] != 'visible'):
@@ -108,12 +110,13 @@ def draw_stacking_context(stream, stacking_context):
 
         if box.style['opacity'] < 1:
             original_stream = stream
-            stream = stream.add_group(stream.page_rectangle)
+            stream = stream.add_group(*stream.page_rectangle)
 
         if box.transformation_matrix:
             if box.transformation_matrix.determinant:
                 stream.transform(*box.transformation_matrix.values)
             else:
+                stream.end_marked_content()
                 return
 
         # Point 1 is done in draw_page
@@ -158,11 +161,15 @@ def draw_stacking_context(stream, stacking_context):
                 if isinstance(block, boxes.ReplacedBox):
                     draw_border(stream, block)
                     draw_replacedbox(stream, block)
-                else:
-                    for child in block.children:
-                        if isinstance(child, boxes.LineBox):
+                elif block.children:
+                    if block != box:
+                        stream.begin_marked_content(block, mcid=True)
+                    if isinstance(block.children[-1], boxes.LineBox):
+                        for child in block.children:
                             draw_inline_level(
                                 stream, stacking_context.page, child)
+                    if block != box:
+                        stream.end_marked_content()
 
             # Point 8
             for child_context in stacking_context.zero_z_contexts:
@@ -183,6 +190,8 @@ def draw_stacking_context(stream, stacking_context):
             stream.draw_x_object(group_id)
             stream.pop_state()
 
+        stream.end_marked_content()
+
 
 def rounded_box_path(stream, radii):
     """Draw the path of the border radius box.
@@ -237,6 +246,8 @@ def draw_background(stream, bg, clip_box=True, bleed=None, marks=()):
         # Background color
         if bg.color.alpha > 0:
             with stacked(stream):
+                stream.set_color_rgb(*bg.color[:3])
+                stream.set_alpha(bg.color.alpha)
                 painting_area = bg.layers[-1].painting_area
                 if painting_area:
                     if bleed:
@@ -249,9 +260,7 @@ def draw_background(stream, bg, clip_box=True, bleed=None, marks=()):
                     stream.rectangle(*painting_area)
                     stream.clip()
                     stream.end()
-                stream.rectangle(*stream.page_rectangle)
-                stream.set_color_rgb(*bg.color[:3])
-                stream.set_alpha(bg.color.alpha)
+                stream.rectangle(*painting_area)
                 stream.fill()
 
         if bleed and marks:
@@ -381,8 +390,8 @@ def draw_background_image(stream, layer, image_rendering):
     matrix = Matrix(e=position_x + positioning_x, f=position_y + positioning_y)
     matrix @= stream.ctm
     pattern = stream.add_pattern(
-        image_width, image_height, repeat_width, repeat_height, matrix)
-    group = pattern.add_group([0, 0, repeat_width, repeat_height])
+        0, 0, image_width, image_height, repeat_width, repeat_height, matrix)
+    group = pattern.add_group(0, 0, repeat_width, repeat_height)
 
     with stacked(stream):
         layer.image.draw(group, image_width, image_height, image_rendering)
@@ -415,6 +424,17 @@ def draw_border(stream, box):
     # We need a plan to draw beautiful borders, and that's difficult, no need
     # to lie. Let's try to find the cases that we can handle in a smart way.
 
+    def get_columns_with_rule():
+        """Yield columns that have a rule drawn on the left."""
+        skip_next = True
+        for child in box.children:
+            if child.style['column_span'] == 'all':
+                skip_next = True
+            elif skip_next:
+                skip_next = False
+            else:
+                yield child
+
     def draw_column_border():
         """Draw column borders."""
         columns = (
@@ -423,14 +443,14 @@ def draw_border(stream, box):
                 box.style['column_count'] != 'auto'))
         if columns and box.style['column_rule_width']:
             border_widths = (0, 0, 0, box.style['column_rule_width'])
-            for child in box.children[1:]:
+            for child in get_columns_with_rule():
                 with stacked(stream):
                     position_x = (child.position_x - (
                         box.style['column_rule_width'] +
                         box.style['column_gap']) / 2)
                     border_box = (
                         position_x, child.position_y,
-                        box.style['column_rule_width'], box.height)
+                        box.style['column_rule_width'], child.height)
                     clip_border_segment(
                         stream, box.style['column_rule_style'],
                         box.style['column_rule_width'], 'left', border_box,
@@ -517,7 +537,7 @@ def clip_border_segment(stream, style, width, side, border_box,
         pi" Quart. J. Pure. Appl. Math., vol. 45 (1913-1914), pp. 350-372],
         wonderfully explained by Dr Rob.
 
-        http://mathforum.org/dr.math/faq/formulas/
+        https://mathforum.org/dr.math/faq/formulas/
 
         """
         x = (a - b) / (a + b)
@@ -666,37 +686,38 @@ def clip_border_segment(stream, style, width, side, border_box,
 
 
 def draw_rounded_border(stream, box, style, color):
-    rounded_box_path(stream, box.rounded_padding_box())
     if style in ('ridge', 'groove'):
-        rounded_box_path(stream, box.rounded_box_ratio(1 / 2))
         stream.set_color_rgb(*color[0][:3])
         stream.set_alpha(color[0][3])
-        stream.fill(even_odd=True)
+        rounded_box_path(stream, box.rounded_padding_box())
         rounded_box_path(stream, box.rounded_box_ratio(1 / 2))
-        rounded_box_path(stream, box.rounded_border_box())
+        stream.fill(even_odd=True)
         stream.set_color_rgb(*color[1][:3])
         stream.set_alpha(color[1][3])
+        rounded_box_path(stream, box.rounded_box_ratio(1 / 2))
+        rounded_box_path(stream, box.rounded_border_box())
         stream.fill(even_odd=True)
         return
+    stream.set_color_rgb(*color[:3])
+    stream.set_alpha(color[3])
+    rounded_box_path(stream, box.rounded_padding_box())
     if style == 'double':
         rounded_box_path(stream, box.rounded_box_ratio(1 / 3))
         rounded_box_path(stream, box.rounded_box_ratio(2 / 3))
     rounded_box_path(stream, box.rounded_border_box())
-    stream.set_color_rgb(*color[:3])
-    stream.set_alpha(color[3])
     stream.fill(even_odd=True)
 
 
 def draw_rect_border(stream, box, widths, style, color):
     bbx, bby, bbw, bbh = box
     bt, br, bb, bl = widths
-    stream.rectangle(*box)
     if style in ('ridge', 'groove'):
+        stream.set_color_rgb(*color[0][:3])
+        stream.set_alpha(color[0][3])
+        stream.rectangle(*box)
         stream.rectangle(
             bbx + bl / 2, bby + bt / 2,
             bbw - (bl + br) / 2, bbh - (bt + bb) / 2)
-        stream.set_color_rgb(*color[0][:3])
-        stream.set_alpha(color[0][3])
         stream.fill(even_odd=True)
         stream.rectangle(
             bbx + bl / 2, bby + bt / 2,
@@ -706,6 +727,9 @@ def draw_rect_border(stream, box, widths, style, color):
         stream.set_alpha(color[1][3])
         stream.fill(even_odd=True)
         return
+    stream.set_color_rgb(*color[:3])
+    stream.set_alpha(color[3])
+    stream.rectangle(*box)
     if style == 'double':
         stream.rectangle(
             bbx + bl / 3, bby + bt / 3,
@@ -714,8 +738,6 @@ def draw_rect_border(stream, box, widths, style, color):
             bbx + bl * 2 / 3, bby + bt * 2 / 3,
             bbw - (bl + br) * 2 / 3, bbh - (bt + bb) * 2 / 3)
     stream.rectangle(bbx + bl, bby + bt, bbw - bl - br, bbh - bt - bb)
-    stream.set_color_rgb(*color[:3])
-    stream.set_alpha(color[3])
     stream.fill(even_odd=True)
 
 
@@ -936,10 +958,15 @@ def draw_inline_level(stream, page, box, offset_x=0, text_overflow='clip',
         draw_background(stream, box.background)
         draw_border(stream, box)
         if isinstance(box, (boxes.InlineBox, boxes.LineBox)):
+            link_annotation = None
             if isinstance(box, boxes.LineBox):
                 text_overflow = box.text_overflow
                 block_ellipsis = box.block_ellipsis
+            else:
+                link_annotation = box.link_annotation
             ellipsis = 'none'
+            if link_annotation:
+                stream.begin_marked_content(box, mcid=True, tag='Link')
             for i, child in enumerate(box.children):
                 if i == len(box.children) - 1:
                     # Last child
@@ -956,6 +983,8 @@ def draw_inline_level(stream, page, box, offset_x=0, text_overflow='clip',
                     draw_inline_level(
                         stream, page, child, child_offset_x, text_overflow,
                         ellipsis)
+            if link_annotation:
+                stream.end_marked_content()
         elif isinstance(box, boxes.InlineReplacedBox):
             draw_replacedbox(stream, box)
         else:
@@ -1072,22 +1101,21 @@ def draw_first_line(stream, textbox, text_overflow, block_ellipsis, x, y,
     utf8_text = textbox.pango_layout.text.encode()
     previous_utf8_position = 0
 
-    runs = [first_line.runs[0]]
-    while runs[-1].next != ffi.NULL:
-        runs.append(runs[-1].next)
-
     matrix = Matrix(1, 0, 0, -1, x, y)
     if angle:
-        matrix = Matrix(a=cos(angle), b=-sin(angle),
-                        c=sin(angle), d=cos(angle)) @ matrix
+        a, c = cos(angle), sin(angle)
+        b, d = -c, a
+        matrix = Matrix(a, b, c, d) @ matrix
     stream.text_matrix(*matrix.values)
     last_font = None
     string = ''
     x_advance = 0
     emojis = []
-    for run in runs:
+    run = first_line.runs[0]
+    while run != ffi.NULL:
         # Pango objects
-        glyph_item = ffi.cast('PangoGlyphItem *', run.data)
+        glyph_item = run.data
+        run = run.next
         glyph_string = glyph_item.glyphs
         glyphs = glyph_string.glyphs
         num_glyphs = glyph_string.num_glyphs
@@ -1107,8 +1135,8 @@ def draw_first_line(stream, textbox, text_overflow, block_ellipsis, x, y,
             if string:
                 stream.show_text(string)
             string = ''
+            stream.set_font_size(font.hash, 1 if font.bitmap else font_size)
             last_font = font
-        stream.set_font_size(font.hash, 1 if font.bitmap else font_size)
         string += '<'
         for i in range(num_glyphs):
             glyph_info = glyphs[i]
@@ -1129,19 +1157,6 @@ def draw_first_line(stream, textbox, text_overflow, block_ellipsis, x, y,
             if glyph not in font.widths:
                 pango.pango_font_get_glyph_extents(
                     pango_font, glyph, stream.ink_rect, stream.logical_rect)
-                x1, y1, x2, y2 = (
-                    stream.ink_rect.x,
-                    -stream.ink_rect.y - stream.ink_rect.height,
-                    stream.ink_rect.x + stream.ink_rect.width,
-                    -stream.ink_rect.y)
-                if x1 < font.bbox[0]:
-                    font.bbox[0] = int(units_to_double(x1 * 1000) / font_size)
-                if y1 < font.bbox[1]:
-                    font.bbox[1] = int(units_to_double(y1 * 1000) / font_size)
-                if x2 > font.bbox[2]:
-                    font.bbox[2] = int(units_to_double(x2 * 1000) / font_size)
-                if y2 > font.bbox[3]:
-                    font.bbox[3] = int(units_to_double(y2 * 1000) / font_size)
                 font.widths[glyph] = int(round(
                     units_to_double(stream.logical_rect.width * 1000) /
                     font_size))
